#!/usr/bin/perl -w
# mysar.pl - see header for version
# Record mysql status and variables in a DB table
# $Header$
# 
# Updated by Pythian, Inc. 2009
# www.pythian.com
# 
# Copyright (C) 2009 Gerardo Narvaja
#
# This utility is loosely based on mysqltuner and statpack
# This code is intended to ultimately be included on a new utility
# being developed @ Pythian
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# This project would not be possible without help from:
#	Sheeri Cabral
#	Tim Procter
#
# Project page: http://launchpad.net/mysar

use strict;
use warnings;
use diagnostics;
use Getopt::Long;
use DBI;

# dummy variable to eliminate code while debugging
my $bigassif = 1;

# Set up a few variables for use in the script
my $mysarversion = "0.3-alpha";

# Set defaults
my %opt = (
		"host" => 0,
		"socket" => 0,
		"port" => 0,
		"user" => 0,
		"pass" => 0,
		"variables" => 0,
		"status" => 0,
		"varfile" => 0,
		"statfile" => 0,
		"database" => 0,
	);
	
# Gather the options from the command line
GetOptions(\%opt,
		'host=s',
		'socket=s',
		'port=i',
		'user=s',
		'pass=s',
		'database=s',
		'variables',
		'status',
		'varfile=s',
		'statfile=s',
		'help'
	);

if (defined $opt{'help'} && $opt{'help'} == 1) { usage(); }

sub usage {
	# Shown when --help option is passed
	print "\n".
		"   mysar $mysarversion - MySQL Variables gathering utility modeled after 'sar'\n".
		"   Maintained by Gerardo Narvaja - Licensed under GPL v2\n".
		"\n".
		"   Options List:\n".
		"      --host <hostname>    Connect to a remote host to perform tests (default: localhost)\n".
		"      --socket <socket>    Use a different socket for a local connection\n".
		"      --port <port>        Port to use for connection (default: 3306)\n".
		"      --user <username>    (required) Username to use for authentication\n".
		"      --pass <password>    (required) Password to use for authentication\n".
		"      --variables          Gather the global variables values\n".
		"      --status             Gather the global status values\n".
		"                           You need to specify at least one of the options above\n".
		"      --varfile            File with the list of variables to monitor. (default: all)\n".
		"      --statfile           File with the list of status variables to monitor (default: all)\n".
		"                           For variables/status specification you may use '%' as wildcard\n".
		"      --database           Specifies which schema contains the tables 'value_var' & 'value_stat'\n". 
		"                           (default: avail)\n".
		"\n";
	exit;
}

# DB variables
my $dsn = "DBI:mysql";			# connection string seed
my $dbh;				# DB  handler
my $sth;				# DB stat
my $my_database = "avail";		# Which database schema to use

sub mysql_setup {
	my $my_user = "";
	my $my_pass = "";
	my $my_host = "localhost";
	my $my_socket = 0;
	my $my_port = "3306";

	# Pick the connection parameters
	# user and password are mandatory
	if ($opt{user} ne 0) {
		$my_user = "$opt{user}";
		print "User: $my_user\n";
	} else {
		print "ERROR: You need to specify user name w/ --user\n";
		usage();
		exit 1;
	}
	if ($opt{pass} ne 0) {
		$my_pass = "$opt{pass}";
	} else { 
		print "ERROR: You need to specify user password w/ --pass\n";
		usage();
		exit 1;
	}
	
	# Optional connection parameters
	if ($opt{database} ne 0) {
		$my_database = "$opt{database}";
		print "Using tables $my_database.value_var & $my_database.value_stat\n";
	}
	if ($opt{host} ne 0) {
		$my_host = "$opt{host}";
		print "The values are going to be retrieved from: $my_host\n";
	}
	if ($opt{port} ne 0) {
		$my_port = "$opt{port}";
		print "Using port $my_port\n";
	}
	# Use default or specified values
	$dsn = "$dsn:$my_database;host=$my_host;port=$my_port";

	# Test for socket specification
	if ($opt{socket} ne 0) {
		$my_socket = "$opt{socket}";
		$dsn = "$dsn;mysql_socket=$my_socket";
	}
print "dsn: *$dsn*\n";
	
	# Connect to the DB
	$dbh = DBI->connect($dsn, $my_user, $my_pass)
		or die "ERROR: Unable to connect to MySQL server $my_host w/ user: $my_user\n".
                       "   --> $!\n";
}

# Function to check if the tables exist
sub tbl_exists($) {
	my (@tblexists);
	my $table = $_[0];

	eval {
		$dbh->{PrintError} = 0;
		if($dbh->{RaiseError} == 0){
			$dbh->{RaiseError} = 1;
			print "Failed to find table $my_database.$table\n";
        		}

		};
	my $sth = $dbh->prepare("SHOW TABLES LIKE '$table'") or return 0;
	$sth->execute() or return 0;
	@tblexists = $sth->fetchrow_array() or return 0;
	if ($tblexists[0]) { return 1;} else {return 0}
	$sth->finish();

	if ($@) {
		print "returning 0 for $table\n";
        	return 0;
	}
	 print "returning 1 for $table\n";
	return 1;
}

# Verify the tables and create them if they don't exist
sub verify_table($) {
	my $table = $_[0];

	if (tbl_exists($table) eq 1) { return; }

	print "*** Creating $my_database.$table \n";
	if ($table eq "value_var") {
		$dbh->do("CREATE TABLE IF NOT EXISTS `value_var` ( ".
			"`value_var_id` int(11) NOT NULL auto_increment, ".
			"`value_var_timestamp` timestamp NOT NULL default CURRENT_TIMESTAMP, ".
			"`value_var_name` varchar(64) NOT NULL, ".
			"`value_var_number` bigint(20) default NULL, ".
			"`value_var_string` varchar(256) default NULL, ".
			"PRIMARY KEY  (`value_var_id`) ".
			") ENGINE=InnoDB")
				or die "ERROR: Failed to create $table\n";
	}
	if ($table eq "value_stat") {
		$dbh->do("CREATE TABLE IF NOT EXISTS `value_stat` ( ".
			"`value_stat_id` int(11) NOT NULL auto_increment, ".
			"`value_stat_timestamp` timestamp NOT NULL default CURRENT_TIMESTAMP, ".
			"`value_stat_name` varchar(64) NOT NULL, ".
			"`value_stat_number` bigint(20) default NULL, ".
			"`value_stat_delta` bigint(20) default NULL," .
			"`value_stat_string` varchar(256) default NULL, ".
			"PRIMARY KEY  (`value_stat_id`) ".
			") ENGINE=InnoDB ")
				or die "ERROR: Failed to create $table\n";
	}
}


# Flags to determine what to pick from the database
my $my_variables = 0;
my $my_all_variables = 0;
my $my_status = 0;
my $my_all_status = 0;

# Populates all of the variable and status hashes
sub get_all_vars {
	my (@myvarlist_names, @myvarlist_values,@mystatlist_names,@mystatlist_values);
	my ($in_var, $in_stat);
	# We need to initiate at least one query so that our data is useable
	if ($opt{varfile} ne 0) {
		print "Using variables list '$opt{varfile}'\n";
		open($in_var,"< $opt{varfile}") or die "$opt{varfile}: $!\n";
	} else {
		print "Will retrieve all variables\n";
		$my_all_variables = 1;
	}
	if ($opt{'statfile'} ne 0) {
		print "Using status list '$opt{statfile}'\n";
		open($in_stat,"< $opt{statfile}") or die "$opt{statfile}: $!\n";
	} else {
		print "Will retrieve all status variables\n";
		$my_all_status = 1;
	}

	# Retrieve the timestamp straight from the DB
	my (@ts_out, $ts);
	@ts_out = $dbh->selectrow_array("SELECT NOW()")
		or die "ERROR: SELECT NOW() failed\n".
                       "   --> $!\n";
	$ts = $ts_out[0]; print "*** Sampled at: $ts\n";

	# Retrieve the list of (global) variables
	if ($my_variables eq 1) {	# value is set at the begining of the main section
		print "*** Retrieving configuration variables from the DB\n";
		my ($var_name, $var_value);
		if ($my_all_variables eq 1) {
			$sth = $dbh->prepare("SHOW GLOBAL VARIABLES")
				or die "ERROR: SHOW GLOBAL VARIABLES failed\n";
			$sth->execute();
			while ( @myvarlist_values = $sth->fetchrow_array() ) {
				# print "$ts - Row: $myvarlist_values[0] -> $myvarlist_values[1]\n";
				$var_name = $myvarlist_values[0];
				$var_value = $myvarlist_values[1];
				
				if  (($var_value =~ /\D/) || (length($var_value) eq 0)) {
					$dbh->do ("INSERT INTO value_var(value_var_timestamp, value_var_name,".
						" value_var_string) VALUES (\"$ts\", '$var_name', \"$var_value\")")
						or die "ERROR: INSERT for STATUS $var_name ($var_value) failed\n";
				} else {
					$dbh->do ("INSERT INTO value_var(value_var_timestamp, value_var_name,".
						" value_var_number) VALUES (\"$ts\", '$var_name', $var_value)")
						or die "ERROR: INSERT for STATUS $var_name ($var_value) failed\n";
				}
			}
		} else {
			$sth = $dbh->prepare("SHOW GLOBAL VARIABLES LIKE ?")
				or die "ERROR: SHOW GLOBAL VARIABLES failed\n";
			while ( <$in_var> ) {
				chomp; my $var_ref = $_;
				print "***> var names: *$var_ref*\n";
				$sth->execute($var_ref) or die "*** miserably  ***";
				while ( @myvarlist_values = $sth->fetchrow_array() ) {
					# print "$ts - Row: $myvarlist_values[0] -> $myvarlist_values[1]\n";
					$var_name = $myvarlist_values[0];
					$var_value = $myvarlist_values[1];
					
					if  (($var_value =~ /\D/) || (length($var_value) eq 0)) {
						$dbh->do ("INSERT INTO value_var(value_var_timestamp, value_var_name,".
							" value_var_string) VALUES (\"$ts\", '$var_name', \"$var_value\")")
							or die "ERROR: INSERT for STATUS $var_name ($var_value) failed\n";
					} else {
						$dbh->do ("INSERT INTO value_var(value_var_timestamp, value_var_name,".
							" value_var_number) VALUES (\"$ts\", '$var_name', $var_value)")
							or die "ERROR: INSERT for STATUS $var_name ($var_value) failed\n";
					}
				}
			close $in_var;
			}
		}
	}	# $my_variables eq 1

	# Retrieve the list of (global) status
	if ($my_status eq 1) {	# value is set at the begining of the main section
		print "*** Retrieving status variables from the DB\n";
		my ($stat_name, $stat_value);
		if ($my_all_status eq 1) {
			$sth = $dbh->prepare("SHOW GLOBAL STATUS")
				or die "ERROR: SHOW GLOBAL STATUS failed\n";
			$sth->execute();
			while ( @mystatlist_values = $sth->fetchrow_array() ) {
				# print "$ts - Row: $mystatlist_values[0] -> $mystatlist_values[1]\n";
				$stat_name = $mystatlist_values[0];
				$stat_value = $mystatlist_values[1];
				
				if  (($stat_value =~ /\D/) || (length($stat_value) eq 0)) {
					$dbh->do ("INSERT INTO value_stat(value_stat_timestamp, value_stat_name,".
						" value_stat_string) VALUES (\"$ts\", '$stat_name', \"$stat_value\")")
						or die "ERROR: INSERT for STATUS $stat_name ($stat_value) failed\n";
				} else {
					$dbh->do ("INSERT INTO value_stat(value_stat_timestamp, value_stat_name,".
						" value_stat_number) VALUES (\"$ts\", '$stat_name', $stat_value)")
						or die "ERROR: INSERT for STATUS $stat_name ($stat_value) failed\n";
				}
			}
		} else {
			$sth = $dbh->prepare("SHOW GLOBAL STATUS LIKE ?")
				or die "ERROR: SHOW GLOBAL STATUS failed\n";
			while ( <$in_stat> ) {
				chomp; my $stat_ref = $_;
				print "***> status names: *$stat_ref*\n";
				$sth->execute($stat_ref);
				while ( @mystatlist_values = $sth->fetchrow_array() ) {
					# print "$ts - Row: $mystatlist_values[0] -> $mystatlist_values[1]\n";
					$stat_name = $mystatlist_values[0];
					$stat_value = $mystatlist_values[1];
					
					if  (($stat_value =~ /\D/) || (length($stat_value) eq 0)) {
						$dbh->do ("INSERT INTO value_stat(value_stat_timestamp, value_stat_name,".
							" value_stat_string) VALUES (\"$ts\", '$stat_name', \"$stat_value\")")
							or die "ERROR: INSERT for STATUS $stat_name ($stat_value) failed\n";
					} else {
						$dbh->do ("INSERT INTO value_stat(value_stat_timestamp, value_stat_name,".
							" value_stat_number) VALUES (\"$ts\", '$stat_name', $stat_value)")
							or die "ERROR: INSERT for STATUS $stat_name ($stat_value) failed\n";
					}
				}
			}
			close $in_stat;
		}
	}
}

# ---------------------------------------------------------------------------
# BEGIN 'MAIN'
# ---------------------------------------------------------------------------
print	"\n * mysar v$mysarversion - Gerardo Narvaja\n".
	" * Developed at Pythian Inc. http://www.pythian.com\n".
	" * Use \"perldoc mysar.pl\" for usage information";

if ($opt{status} ne 0) { $my_status = 1; }
if ($opt{variables} ne 0) { $my_variables = 1; }
if (($my_variables eq 0) && ($my_status eq 0)) {
	print "\nERROR: You need to specify --status and/or --variables\n\n";
	usage();
	exit 1;
	}

mysql_setup();				# Gotta login first
verify_table("value_var");		# Verify the tables exist, if not create them
verify_table("value_stat");
get_all_vars();				# Get the variables and status and store them into the DB
$dbh->disconnect();			# Disconnect from the DB
print "OK - Finished successfully\n";
exit 0;
# ---------------------------------------------------------------------------
# END 'MAIN'
# ---------------------------------------------------------------------------

# ---------------------------------------------------------------------------
# pod format documentation
# ---------------------------------------------------------------------------

=pod

=head1 NAME

mysar - sar-like utility for MySQL

=head1 VERSION

Current version is v0.3-alpha
Report bugs @ http://bugs.launchpad.net/mysar/

=head1 DESCRIPTION

This utiliy sole purpose is to take a snapshot of MySQL's running status in 
a number of tables so it they can be queried and cross reference with other 
utilities and logs output. For example: sar or MySQL's different logs.

=head1 USAGE 

 Options List:
      --host <hostname>    Connect to a remote host to perform tests (default: localhost)
      --socket <socket>    Use a different socket for a local connection
      --port <port>        Port to use for connection (default: 3306)
      --user <username>    (required) Username to use for authentication
      --pass <password>    (required) Password to use for authentication
      --variables          Gather the global variables values
      --status             Gather the global status values
                           You need to specify at least one of the options above
      --varfile            File with the list of variables to monitor. (default: all)
      --statfile           File with the list of status variables to monitor (default: all)
                           For variables/status specification you may use '%' as wildcard
      --database           Specifies which schema contains the tables 'value_var' & 'value_stat'
                           (default: avail)

=cut
